from subprocess import call, check_output, DEVNULL, CalledProcessError
from os import listdir
from pathlib import Path
from typing import Optional

from .interactive import pr, choose, cyan


def choose_file(root_dir: Path) -> Optional[Path]:
    """
    Interactive loop to choose a file
    """
    def _format(root_dir: Path, entry: str) -> str:
        f = root_dir / entry
        if f.is_dir():
            return entry
        return f'{entry}\t({human_bytes(f.stat().st_size)}, {count_lines(f)})'

    listing = listdir(root_dir)
    if not listing:
        return pr('Empty directory!', '!')
    while 1:
        c = choose([_format(root_dir, i) for i in listing], default=-1)
        if c < 0:
            return
        f = root_dir / listing[c]
        if f.is_dir():
            f = choose_file(root_dir / f)
            if not f:
                continue
        return f


def count_lines(file_path: Path) -> int:
    """ Return count of line endings in a given file """
    return 1 + file_path.read_bytes().count(b'\n')


def human_bytes(size_in_bytes: int) -> str:
    """ Return a human-readable size representation """
    unit = 0
    size_in_bytes = int(size_in_bytes)
    while size_in_bytes >= 1024:
        unit += 1
        size_in_bytes /= 1024
    return str(round(size_in_bytes)) + ('', 'KB', 'MB', 'GB', 'TB')[unit]


def file_volume(path: Path) -> tuple:
    """
    Returns:
        [0] => Size in bytes
        [1] => Lines count
        [2] => A colored string that includes a human-readable representation of both values 
    """
    sb = path.stat().st_size
    lc = count_lines(path)
    return sb, lc, f'{cyan(path.name)} ({human_bytes(sb)}, {lc})'


def is_image(image_path: str) -> (str, None):
    """
    Checks the file signature (magic number)
            for an image

    :param image_path: The path to the image
    :return: True if the image is PNG or JPG
    """

    signatures = {'JPG': 'ffd8ff',
                  'PNG': '89504e',
                  'GIF': '474946'}

    with open(image_path, 'rb') as img_file:
        signature = img_file.read(3).hex()
        for sig in signatures:
            if signature == signatures[sig]:
                return sig
    return None


def check_dependencies(package_names: iter) -> (None, list):
    """
    Check each one of the system packages if its installed
    """
    if not package_names:
        raise ValueError('"package_names" should not be empty')

    return {package: is_package(package) for package in package_names}


def is_package(package_name: str) -> (str, None):
    """
    Check if a system package is installed by accessing the available package manager

    :param package_name: Package to check
    :return: The version of the installed package or None if no such package
    """
    try:
        return check_output(['/usr/bin/pacman', '-Q', package_name], stderr=DEVNULL).decode().strip().split(' ')[1]
    except CalledProcessError:
        return  # No such package
    except FileNotFoundError:
        try:
            return check_output(('apt', 'list', '-qq', package_name)).decode().strip().split(' ')[1]
        except CalledProcessError:
            pass
    return




from re import match
from subprocess import call, check_output, DEVNULL, CalledProcessError

from .interactive import pr


def is_mac(mac: str) -> bool:
    """
    Check if the specified string is a mac address
    Bytes delimiter can be either '-' or ':'
    The MAC string is stripped.
    :param mac: The MAC address string
    :return: True if the MAC address is valid
    """
    return bool(match("[0-9a-f]{2}([-:])[0-9a-f]{2}(\\1[0-9a-f]{2}){4}$", mac.strip().lower()))


def get_net() -> (None, dict):
    # Uses iproute2 (ip)

    # Example raw:
    # default via 192.168.1.1 dev eth0
    # 192.168.1.0/24 dev eth0 proto kernel scope link src 192.168.1.28

    raw = check_output(['ip', 'route']).decode().strip().split('\n')
    d = {}
    if not raw or 'default' not in raw[0]:
        return None  # No route
    else:
        # default: (gateway, interface)
        d.update({'default': raw[0].split(' ')[2:5:2]})

    for li in raw[1:]:
        ls = li.split(' ')
        d.update({ls[2]: {'interface': ls[2], 'subnet': ls[0], 'ip': ls[8]}})

    return d



from time import time

from .interactive import pr


class IterationTimer:
    def __init__(self, total: int, init_interval: int = 1, max_interval: int = 15, interval_amp_mod: float = 2):
        if total < 1:
            raise ValueError('"total" cannot be lower than 1')
        if init_interval < 1:
            raise ValueError('"init_interval" cannot be lower than 1')
        if max_interval < init_interval:
            raise ValueError(
                '"max_interval" cannot be lower than "init_interval"')
        if interval_amp_mod < 1.0:
            raise ValueError('"interval_amp_mod" cannot be lower than 1')

        self.total = total
        self.interval = init_interval
        self.max_interval = max_interval
        self.current = self.last = 0
        self.lt = time()

    def tick(self) -> None:
        # Progress
        c = self.current = self.current + 1
        ts = time()

        if ts - self.lt > self.interval:
            # Show status
            delta = c - self.last
            if not delta:
                return pr('No progression!', '!')
            spd = int(delta / self.interval)
            prcnt = c / (self.total / 100)
            eta = int((self.total - c) / spd)
            pr('%.2f%% ' % prcnt +
               f'[{c}/{self.total}]\t@ {spd} ps\tETA: {eta} secs')

            # Checkpoint
            self.lt = ts
            self.last = c
            if self.interval <= self.max_interval:
                self.interval *= 2



from termcolor import colored, cprint


def clear() -> None:
    """
    Clear the screen by sending an escape sequence
    """
    print(chr(27) + "[2J")


def cyan(text: str) -> str:
    """
    Color a string in cyan
    """
    return colored(text, 'cyan')


def pr(text: str, notation='+', end='\n') -> None:
    """
    Fancy print function
    Adds a colored bracket-wrapped sign in front of the text
    * Default sign is '+'
    """
    col = {
        '+': 'green',
        '*': 'cyan',
        '~': 'cyan',
        'X': 'red',
        '!': 'yellow',
        '?': 'blue'
    }[notation]
    print(f"{colored(f'[{notation}]' , col)} {text}", end=end)


def choose(options: iter = ('Yes', 'No'), prompt: str = 'Choose action:', default: int = 0) -> int:
    """
    Presents a list of options to the user,
        allowing him to numerically select a desired one.

    * Returns: The selected item's index
    * Returns: -2 => KeyboardInterrupt caught
    * Returns: -1 => Bad input received
    """
    if not options:
        raise ValueError(
            " [!] No options passed to choice() !!!")  # No options
    pr(prompt, '?')
    for index, option in enumerate(options):
        line = '\t'
        if index == default:
            line += '[%d]. ' % (index + 1)
        else:
            line += ' %d.  ' % (index + 1)
        line += option
        cprint(line, 'yellow')
    try:
        ans = input(colored('[>>>] ', 'yellow'))
        if not ans:
            return default
        ans = int(ans)
        assert 0 < ans <= len(options)
        return ans - 1
    except KeyboardInterrupt:
        return -2  # Keyboard Interrupt
    except AssertionError:
        return -1  # Bad Number
    except ValueError:
        return -1  # Probably text received


def ask(question: str) -> (None, str, int):
    """
    Ask the user something by giving him a question

    * Returns: The response as int if possible
        otherwise as str, None if no response
    * Expect: KeyboardInterrupt
    """
    pr(question, '?')
    answer = input('>')
    if answer == '':
        return None
    try:
        answer = int(answer)
    except ValueError:
        pass
    return answer


def pause(reason: str = 'continue', cancel: bool = False) -> bool:
    """
    Show a message and wait for an enter key to be pressed

    Returns: True after [ENTER] pressed
    Returns: False after [^C] pressed
    """
    s = 'Press %s to %s' % (colored('[ENTER]', 'cyan'), reason)
    if cancel:
        s += ', %s to cancel' % colored('[^C]', 'red')
    pr(s, '?')

    try:
        input()
        return True
    except KeyboardInterrupt:
        return False


def banner(txt: str, style: str = 'slant') -> str:
    """
    A static wrapper around pyfiglet that allows easy creation of figlets

    Requires: "pyfiglet"
    Param: style: The style (From: /usr/lib/python*/site-packages/pyfiglet/fonts/)
    Returns: The created ASCII art
    """
    try:
        from pyfiglet import Figlet
    except ImportError:
        pr('Module "pyfiglet" not installed, rendering legacy banner', '!')
        return '~=~=~ %s ~=~=~' % txt
    f = Figlet(font=style)
    return f.renderText(text=txt)


def get_date() -> str:
    """
    Returns: Today's date (e.g. "28.11.2017" ;P)
    """
    from datetime import datetime
    return datetime.now().strftime("%d.%m.%Y")



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="interutils",
    version="0.11.2",
    author="Max G",
    author_email="max3227@gmail.com",
    description="A collection of utilities for creating interactive console scripts",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=[
        'termcolor',
    ],
)



from .interactive import clear, pr, ask, banner, choose, pause, cyan, get_date
from .net import get_net, is_mac
from .fs import choose_file, count_lines, file_volume, human_bytes, is_image, is_package, check_dependencies
from .iteration_timer import IterationTimer
